/* libwince project */
#include <dirent.h>

#include <sys/stat.h>
#include <sys/types.h>

#include <unistd.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>

#include "libwince.h"

#define CHR_SEPARATOR '/'
#define STR_SEPARATOR "/"
#define STR_UNSUPPORTED "wince: unsupported"

/* errno */
static int _errno = 0;

/* absolute program path */
static const char* _prog = NULL;

/* environment spool */
struct _env_entry
{
    char* str;
    int free;
    struct _env_entry* next;
};
typedef struct _env_entry _env_entry;
static _env_entry* _env_head = NULL;

void libwince_quit(void);
void to_unix_path(char*);
void libwince_open_logging(void);
void libwince_close_logging(void);

/* --------------------- */

void libwince_init(const char* prog, int debug)
{
    errno = 0;
    _prog = NULL;


    if(prog)
	_prog = prog;

    _env_head = NULL;

    srand(time(0));

    if(debug) libwince_open_logging();

    atexit(libwince_quit);
}

void libwince_quit(void)
{
    /* clear environment */
    clearenv();
    libwince_close_logging();
    _prog = NULL;
}

void to_unix_path(char* path)
{
    while(path && *path){ if(*path == '\\') *path = CHR_SEPARATOR; ++path; }
}

void libwince_open_logging(void)
{
    char path[MAX_PATH];

    getcwd(path, sizeof(path));
    pathcat(path, sizeof(path), "stdout.txt", NULL);
    freopen(path, "w", stdout);

    getcwd(path, sizeof(path));
    pathcat(path, sizeof(path), "stderr.txt", NULL);
    freopen(path, "w", stderr);
}

void libwince_close_logging(void)
{
    fclose(stdout);
    fclose(stderr);

    FILE* file = NULL;
    char path[MAX_PATH];

    getcwd(path, sizeof(path));
    pathcat(path, sizeof(path), "stdout.txt", NULL);
    file = fopen(path, "rb");
    if(file)
    {
	int empty = fgetc(file) == EOF ? 1 : 0;
        fclose(file);
	if (empty) unlink(path);
    }

    getcwd(path, sizeof(path));
    pathcat(path, sizeof(path), "stderr.txt", NULL);
    file = fopen(path, "rb");
    if(file)
    {
	int empty = fgetc(file) == EOF ? 1 : 0;
        fclose(file);
	if (empty) unlink(path);
    }
}

int* get_errno(void)
{
    return &_errno;
}

char* get_current_dir_name(void)
{
    if(_prog)
    {
	char* r = strdup(_prog);
	to_unix_path(r);
	return dirname(r);
    }
    return NULL;
}

char* getcwd(char* buf, size_t size)
{
    if(buf)
    {
	char* d = get_current_dir_name();
	if(strlen(d) <= size)
	    strcpy(buf, d);
	free(d);
    }
    return buf;
}

int is_directory(const char* path)
{
    struct stat fs;
    return path && 0 == stat(path, &fs) && S_ISDIR(fs.st_mode) ? 1 : 0;
}

int is_file(const char* path)
{
    struct stat fs;
    return path && 0 == stat(path, &fs) && S_ISREG(fs.st_mode) ? 1 : 0;
}

int pathcat(char* dst, size_t maxlen, const char* src, ...)
{
    const char** ptr = &src;
    int len = maxlen - strlen(dst) - 1;

    for(; *ptr != NULL && len > 0; ++ptr)
    {
        size_t rsz = strlen(dst);
        if((*ptr)[0] != CHR_SEPARATOR || (0 < rsz && dst[rsz] != CHR_SEPARATOR))
        {
            strncat(dst, STR_SEPARATOR, len);
            len -= strlen(STR_SEPARATOR);
        }                                                                                                             
        if(0 < len)
        {
            strncat(dst, *ptr, len);
            len -= strlen(*ptr);
        }
    }
    to_unix_path(dst);
    return len;
}

char *dirname(char *p)
{
    to_unix_path(p);

    if(p && *p)
    {
	char *c = strrchr(p, CHR_SEPARATOR);

        /* "usr"      "." */
	if(!c)
	{
	    strcpy(p, ".");
	}
	else
        /* "/"        "/" */
        /* "/usr"     "/" */
	if(c == p)
	{
	    strcpy(p, "/");
	}
	else
        /* "/usr/"    "/" */
	if(*(c + 1) == 0)
	{
	    *c = 0;
	    return dirname(p);
	}
        /* "/usr/lib"    "/usr" */
	else
	    *c = 0;
    }
    return p;
}

char *basename(char *p)
{
    to_unix_path(p);

    if(p && *p)
    {
	char *c = strrchr(p, CHR_SEPARATOR);

        /* "usr"     "usr" */
	if(!c) return p;
	else
	if(*(c + 1) == 0)
	{
    	    /* "/"       "/"   */
	    if(c == p)
		return p;
	    /* "/usr/"   "usr" */
	    else
	    {
		*c = 0;
		return basename(p);
	    }
	}
	/* "/usr"    "usr" */
	else
	{
	    char* s = p;
	    c = c + 1;
	    while(*c)
		*s++ = *c++;
	    *(s + 1) = 0;
	}
    }
    return p;
}

int chdir(const char *path)
{
    return 1;
}

int putenv(const char* str)
{
    _env_entry* next = NULL;
    size_t len = 0;
    const char* f = NULL;

    if(!str) return -1;

    f = strchr(str, '=');
    if(!f) return -1;

    len = f - str;

    /* find */
    for(next = _env_head; next != NULL; next = next->next)
	if(next->str && 0 == strncasecmp(next->str, str, len)) break;

    /* if exist */
    if(next)
    {
	if(next->str && next->free) free(next->str);
	next->str = (char*) str;
	next->free = 0;
    }
    else
    {
	next = malloc(sizeof(_env_entry));
	if(!next) return -1;

	next->next = _env_head;
	next->str = (char*) str;
	next->free = 0;
	_env_head = next;
    }

    return 0;
}

char* getenv(const char* str)
{
    _env_entry* next = NULL;
    size_t len = 0;

    if(!str) return NULL;

    len = strlen(str);

    /* find */
    for(next = _env_head; next != NULL; next = next->next)
	if(next->str && 0 == strncasecmp(next->str, str, len)) break;

    return next ? strchr(next->str, '=') + 1 : NULL;
}

int clearenv(void)
{
    _env_entry* next = NULL;

    while(_env_head)
    {
	next = _env_head->next;
	if(_env_head->str && _env_head->free) free(_env_head->str);
	free(_env_head);
	_env_head = next;
    }

    return 0;
}

int setenv(const char* name, const char* value, int overwrite)
{
    _env_entry* next = NULL;
    size_t len = 0;

    if(!name) return -1;

    len = strlen(name);

    /* forward traversal */
    for(next = _env_head; next != NULL; next = next->next)
	if(next->str && 0 == strncasecmp(next->str, name, len)) break;

    /* if exist */
    if(next && overwrite)
    {
	if(next->str && next->free) free(next->str);
	next->str = (char*) malloc(len + strlen(value) + 2);
	if(!next->str) return -1;

	strcpy(next->str, name);
	strcat(next->str, "=");
	strcat(next->str, value);
	next->free = 1;
    }
    else
    {
	next = malloc(sizeof(_env_entry));
	if(!next) return -1;

	next->str = (char*) malloc(len + strlen(value) + 2);
	if(!next->str) return -1;

	next->next = _env_head;
	strcpy(next->str, name);
	strcat(next->str, "=");
	strcat(next->str, value);
	next->free = 1;
	_env_head = next;
    }

    return 0;
}

int unsetenv(const char* name)
{
    _env_entry* prev = NULL;
    _env_entry* next = NULL;
    size_t len = 0;

    if(!name) return -1;

    len = strlen(name);

    /* find prev */
    for(next = _env_head; next != NULL; next = next->next)
	if(next->str && 0 == strncasecmp(next->str, name, len)) break;

    if(next)
    {
	if(next == _env_head)
	{
	    _env_head = NULL;
	}
	else
	{
	    for(prev = _env_head; prev != NULL; prev = next->next)
	    if(prev->next == next) break;
	    prev->next = next->next;
	}

	if(next->str && next->free) free(next->str);
	free(next);
    }

    return 0;
}

LPWSTR cstr_to_LPWSTR(const char* str)
{
    if(str)
    {
	int strsize = MultiByteToWideChar(CP_ACP, 0, str, -1, NULL, 0);
	if(strsize)
	{
    	    LPWSTR lpw = calloc(strsize, sizeof(wchar_t));
	    MultiByteToWideChar(CP_ACP, 0, str, -1, lpw, strsize);
	    return lpw;
	}
    }
    return NULL;
}

char* LPWSTR_to_cstr(const LPWSTR lpw)
{
    if(lpw)
    {
	int lpwsize = WideCharToMultiByte(CP_ACP, 0, lpw, -1, NULL, 0, NULL, NULL);
	if(lpwsize)
	{
	    char* str = (char*) calloc(lpwsize, sizeof(char));
	    WideCharToMultiByte(CP_ACP, 0, lpw, -1, str, lpwsize, NULL, NULL);
    	    return str;
	}
    }
    return NULL;
}

wchar_t* char2wchar(const char* str)
{
    return cstr_to_LPWSTR(str);
}

char* wchar2char(const wchar_t* lpw)
{
    return LPWSTR_to_cstr((const LPWSTR) lpw);
}

int system(const char* command)
{
    LPWSTR lpw = cstr_to_LPWSTR(command);
    int res = -1;

    if(lpw)
    {
	res = CreateProcess(NULL, lpw, NULL, NULL, 0, 0, NULL, NULL, NULL, NULL) ? 0 : -1;
	free(lpw);
    }

    return res;
}

char* setlocale(int category, const char* locale)
{
    return NULL; //STR_UNSUPPORTED;
}

const char* strerror(int errnum)
{
    return STR_UNSUPPORTED;
}

FILE* tmpfile(void)
{
    const char* name = tmpnam(NULL);
    FILE* f = fopen(name, "r+b");
    if(f)
	unlink(name);
    return f;
}

char* tmpnam(char* str)
{
    static char _name[PATH_MAX];

    char* cwd = get_current_dir_name();
    char min = 0x30; /* 0 */
    char max = 0x39; /* 9 */
    char rnd[9];
    struct stat fs;
    int ii;

    if(strlen(cwd) + strlen(STR_SEPARATOR) + 8 + 1 + 3 + 1 < PATH_MAX)
    {
	strcpy(_name, cwd);
	strcat(_name, STR_SEPARATOR);

	for(ii = 0; ii < 8; ++ii)
	rnd[ii] = min + (max - min + 1) * (rand() / (RAND_MAX + 1.0));
	rnd[8] = 0;

	strcat(_name, rnd);
	strcat(_name, ".tmp");

	if(0 == stat(_name, &fs)) return tmpnam(str);
	if(str) strcpy(str, _name);

	return _name;
    }
    free(cwd);

    return NULL;
}

clock_t clock(void)
{
    SYSTEMTIME stCurrent;
    FILETIME   ftCurrent;
    GetSystemTime(&stCurrent);
    return (clock_t) (SystemTimeToFileTime(&stCurrent, &ftCurrent) ? *(clock_t*) &ftCurrent : -1);
}

void GetTime(struct timeval* val)
{
    gettimeofday(val, NULL);
}

long GetInterval(struct timeval* val1, struct timeval* val2)
{
    long res = (val2->tv_sec - val1->tv_sec) * 1000 + (val2->tv_usec - val1->tv_usec) / 1000;
    return res;
}

HMODULE LoadLibraryA(const char *filename)
{
    HMODULE ret;
    wchar_t *wfilename = char2wchar(filename);
    ret = LoadLibraryW(wfilename);
    free(wfilename);
    return ret;
}

FARPROC GetProcAddressA(HINSTANCE hModule, const char* procName)
{
    FARPROC ret;
    wchar_t *wprocname = char2wchar(procName);
    ret = GetProcAddressW(hModule, wprocname);
    free(wprocname);
    return ret;
}

DWORD FormatMessageA(unsigned long dwFlags, LPCVOID lpSource, DWORD dwMessageId,
    DWORD dwLanguageId, LPSTR lpBuffer, DWORD nSize, va_list *Arguments)
{
    DWORD ret = 0;
    wchar_t *wbuf = (wchar_t*) malloc(nSize * sizeof(wchar_t));
    ret = FormatMessageW(dwFlags, lpSource, dwMessageId, dwLanguageId, wbuf, nSize, Arguments);
    WideCharToMultiByte(CP_ACP, 0, wbuf, -1, lpBuffer, nSize, NULL, NULL);
    free(wbuf);
    return ret;
}

DWORD GetModuleFileNameA(HMODULE hModule, LPSTR lpFilename, DWORD nSize)
{
    DWORD ret = 0;
    char *filename = NULL;
    wchar_t wbuf[MAX_PATH + 1] = {0};

    ret = GetModuleFileNameW(hModule, wbuf, nSize);
    if(ret > 0)
    {
	filename = wchar2char(wbuf);
	strncpy(lpFilename, filename, ret);
	free(filename);
    }

   return ret;
}

int GetLocaleInfoA(LCID lcid, LCTYPE lctype, LPSTR str, int len)
{
    wchar_t *wbuf = (wchar_t*) malloc(len * sizeof(wchar_t));
    int ret = GetLocaleInfoW(lcid, lctype, wbuf, len);
    WideCharToMultiByte(CP_ACP, 0, wbuf, -1, str, len, NULL, NULL);
    free(wbuf);
    return ret;
}

int remove(const char *pathname)
{
    if(is_directory(pathname))
	return rmdir(pathname);

    return unlink(pathname);
}

/* directory stream implementation */

struct DIR_CE
{
    HANDLE handle;
    wchar_t *wname;
    struct dirent de;
};

DIR_CE* wince_opendir(const char* filename)
{
    if(NULL== filename)
	return NULL;

    if(strlen(filename) > MAX_PATH)
    {
	fprintf(stderr, "%s: path limit: %s\n", __FUNCTION__, filename);
	return NULL;
    }

    struct stat fs;

    if(stat(filename, &fs) < 0 || !S_ISDIR(fs.st_mode))
    {
	fprintf(stderr, "%s: not directory: %s\n", __FUNCTION__, filename);
	return NULL;
    }

    DIR_CE* dirp = (DIR_CE*) malloc(sizeof(DIR_CE));
    if(!dirp)
    {
	fprintf(stderr, "%s: not enough memory\n", __FUNCTION__);
	return NULL;
    }

    /* create: filename + SEPARATOR + '*' */
    size_t len = strlen(filename) + 3;
    char* str = calloc(len, 1);
    if(!str)
    {
	free(dirp);
	fprintf(stderr, "%s: not enough memory\n", __FUNCTION__);
	return NULL;
    }
    pathcat(str, len, filename, "*", NULL);

    /**/
    dirp->wname = char2wchar(str);
    if(!dirp->wname)
    {
	free(dirp);
	fprintf(stderr, "%s: not enough memory\n", __FUNCTION__);
	return NULL;
    }
    free(str);

    dirp->handle = INVALID_HANDLE_VALUE;

    return dirp;
}

int wince_closedir(DIR_CE* dirp)
{
    if(dirp)
    {
	if(dirp->handle != INVALID_HANDLE_VALUE)
    	    FindClose(dirp->handle);
	free(dirp->wname);
	free(dirp);
    }
    return 1;
}

struct dirent* wince_readdir(DIR_CE* dirp)
{
    if(NULL == dirp)
	return NULL;

    WIN32_FIND_DATAW wFindData;

    if(dirp->handle == INVALID_HANDLE_VALUE)
	dirp->handle = FindFirstFileW(dirp->wname, &wFindData);
    else
    if(! FindNextFileW(dirp->handle, &wFindData))
	dirp->handle = INVALID_HANDLE_VALUE;

    if(dirp->handle == INVALID_HANDLE_VALUE)
    {
	DWORD err = GetLastError();

	switch(err)
	{
	    case ERROR_NO_MORE_FILES:
		//fprintf(stderr, "%s: not more files\n", __FUNCTION__);
		break;

	    case ERROR_FILE_NOT_FOUND:
		fprintf(stderr, "%s: file not found\n", __FUNCTION__);
		break;

	    case ERROR_PATH_NOT_FOUND:
		fprintf(stderr, "%s: path not found\n", __FUNCTION__);
		break;

	    case ERROR_NOT_ENOUGH_MEMORY:
		fprintf(stderr, "%s: not enough memory\n", __FUNCTION__);
		break;

	    default:
		fprintf(stderr, "%s: unknown error\n", __FUNCTION__);
		break;
	}

	return NULL;
    }

    char* str = wchar2char(wFindData.cFileName);

    if(!str)
    {
	fprintf(stderr, "%s: not enough memory\n", __FUNCTION__);
	return NULL;
    }

    dirp->de.d_ino = 0;
    dirp->de.d_reclen = 0;
    strncpy(dirp->de.d_name, str, FILENAME_MAX);
    dirp->de.d_namlen = strlen(dirp->de.d_name);
    free(str);

    return &dirp->de;
}

long wince_telldir(DIR_CE* dirp)
{
    fprintf(stderr, "%s: not implemented\n", __FUNCTION__);
    return 0;
}

void wince_seekdir(DIR_CE* dirp, long offset)
{
    fprintf(stderr, "%s: not implemented\n", __FUNCTION__);
}

void wince_rewinddir(DIR_CE* dirp)
{
    if(dirp && dirp->handle != INVALID_HANDLE_VALUE)
    {
	FindClose(dirp->handle);
	dirp->handle = INVALID_HANDLE_VALUE;
    }
}
